"""Code to carry out fictitious learning"""
import numpy as np


def get_best_response_to_play_count(A, play_count):
    """
    Returns the best response to a belief based on the playing distribution of the opponent
    """
    utilities = A @ play_count
    return np.random.choice(
        np.argwhere(utilities == np.max(utilities)).transpose()[0]
    )


def update_play_count(play_count, play):
    """
    Update a belief vector with a given play
    """
    extra_play = np.zeros(play_count.shape)
    extra_play[play] = 1
    return play_count + extra_play


def fictitious_play(A, B, iterations, play_counts=None):
    """
    Implement fictitious play
    """
    if play_counts is None:
        play_counts = [
            np.array([0 for _ in range(dimension)]) for dimension in A.shape
        ]

    yield play_counts

    for repetition in range(iterations):

        plays = [
            get_best_response_to_play_count(matrix, play_count)
            for matrix, play_count in zip((A, B.transpose()), play_counts[::-1])
        ]

        play_counts = [
            update_play_count(play_count, play)
            for play_count, play in zip(play_counts, plays)
        ]
        yield play_counts



"""
A class for integer pivoting. Used for an implementation of the lemke_howson_lex
algorithm.
"""
import warnings

import numpy as np


def find_pivot_row_lex(tableau, column_index, slack_variables):

    """
    Find the index of the row to pivot.

    Identifies the row to pivot by finding the first row in a lexicographical
    ordering of rows. (First checks minimum ratio test, then uses lexicographical
    ordering to break ties)

    Lexicographical ordering implementation is described in pg 20 of the text below:
    B. von Stengel (2007), Equilibrium computation for two-player games in strategic
    and extensive form. Chapter 3, Algorithmic Game Theory, eds. N. Nisan, T.
    Roughgarden, E. Tardos, and V. Vazirani, Cambridge Univ. Press, Cambridge, 53-78.
    http://www.maths.lse.ac.uk/personal/stengel/TEXTE/agt-stengel.pdf

    C describes the transformations on the system, as stored by the coefficients of
    the slack variables (which is initially the identity matrix). This is required in
    order to keep track of lexicographical ordering after pivoting.

    Cq is the rightmost column of the tableau, used in the minimum ration test.
    """

    # gets correct lexicographical ordering
    C = tableau[:, slack_variables]
    lex_order_reversed = np.lexsort(np.rot90(C))
    lex_order = -lex_order_reversed + lex_order_reversed.shape[0]

    # gets ratio of each row
    pivot_column = tableau[:, column_index]
    Cq = tableau[:, -1]

    # catch divide by zero warning
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            r"invalid value encountered in true_divide|divide by zero encountered in true_divide",
        )

        ratio = np.divide(Cq, pivot_column)

    # filters for column coefficients <=0 (to preserve feasibility)
    filtered_ratio = np.where(
        pivot_column <= 0, np.full(ratio.shape, np.inf), ratio
    )

    return np.lexsort(np.flipud((filtered_ratio, lex_order)))[0]


def find_entering_variable(tableau, pivot_row_index, non_basic_variables):
    """
    Finds the non-basic varible which becomes basic after pivoting
    """

    basic_variables = set(range(tableau.shape[1] - 1)) - non_basic_variables
    for i in basic_variables:
        if tableau[pivot_row_index, i] != 0:
            return i


def pivot_tableau_lex(
    tableau, column_index, slack_variables, non_basic_variables
):
    """
    Pivots the tableau and returns the dropped label
    """
    original_labels = non_basic_variables
    pivot_row_index = find_pivot_row_lex(tableau, column_index, slack_variables)
    pivot_element = tableau[pivot_row_index, column_index]
    entering_variable = find_entering_variable(
        tableau, pivot_row_index, original_labels
    )

    for i, _ in enumerate(tableau):
        if i != pivot_row_index:
            tableau[i, :] = (
                tableau[i, :] * pivot_element
                - tableau[pivot_row_index, :] * tableau[i, column_index]
            )

    return entering_variable



"""A class for a normal form game"""
import numpy as np

from .algorithms.lemke_howson import lemke_howson
from .algorithms.support_enumeration import support_enumeration
from .algorithms.vertex_enumeration import vertex_enumeration
from .learning.fictitious_play import fictitious_play
from .learning.replicator_dynamics import (
    asymmetric_replicator_dynamics,
    replicator_dynamics,
)
from .learning.stochastic_fictitious_play import stochastic_fictitious_play


class Game:
    """
    A class for a normal form game.

    Parameters
    ----------

        - A, B: 2 dimensional list/arrays representing the payoff matrices for
          non zero sum games.
        - A: 2 dimensional list/array representing the payoff matrix for a
          zero sum game.
    """

    def __init__(self, *args):
        if len(args) == 2:
            if (not len(args[0]) == len(args[1])) or (
                not len(args[0][0]) == len(args[1][0])
            ):
                raise ValueError("Unequal dimensions for matrices A and B")
            self.payoff_matrices = tuple([np.asarray(m) for m in args])
        if len(args) == 1:
            self.payoff_matrices = np.asarray(args[0]), -np.asarray(args[0])
        self.zero_sum = np.array_equal(
            self.payoff_matrices[0], -self.payoff_matrices[1]
        )

    def __repr__(self):
        if self.zero_sum:
            tpe = "Zero sum"
        else:
            tpe = "Bi matrix"
        return """{} game with payoff matrices:

Row player:
{}

Column player:
{}""".format(
            tpe, *self.payoff_matrices
        )

    def __getitem__(self, key):
        row_strategy, column_strategy = key
        return np.array(
            [
                np.dot(row_strategy, np.dot(m, column_strategy))
                for m in self.payoff_matrices
            ]
        )

    def vertex_enumeration(self):
        """
        Obtain the Nash equilibria using enumeration of the vertices of the best
        response polytopes.

        Algorithm implemented here is Algorithm 3.5 of [Nisan2007]_.

        1. Build best responses polytopes of both players
        2. For each vertex pair of both polytopes
        3. Check if pair is fully labelled
        4. Return the normalised pair

        Returns
        -------

            equilibria: A generator.
        """
        return vertex_enumeration(*self.payoff_matrices)

    def support_enumeration(self, non_degenerate=False, tol=10 ** -16):
        """
        Obtain the Nash equilibria using support enumeration.

        Algorithm implemented here is Algorithm 3.4 of [Nisan2007]_.

        1. For each k in 1...min(size of strategy sets)
        2. For each I,J supports of size k
        3. Solve indifference conditions
        4. Check that have Nash Equilibrium.

        Returns
        -------

            equilibria: A generator.
        """
        return support_enumeration(
            *self.payoff_matrices, non_degenerate=non_degenerate, tol=tol
        )

    def lemke_howson_enumeration(self):
        """
        Obtain Nash equilibria for all possible starting dropped labels
        using the lemke howson algorithm. See `Game.lemke_howson` for more
        information.

        Note: this is not guaranteed to find all equilibria.

        Returns
        -------

            equilibria: A generator
        """
        for label in range(sum(self.payoff_matrices[0].shape)):
            yield self.lemke_howson(initial_dropped_label=label)

    def lemke_howson(self, initial_dropped_label):
        """
        Obtain the Nash equilibria using the Lemke Howson algorithm implemented
        using integer pivoting.

        Algorithm implemented here is Algorithm 3.6 of [Nisan2007]_.

        1. Start at the artificial equilibrium (which is fully labeled)
        2. Choose an initial label to drop and move in the polytope for which
           the vertex has that label to the edge
           that does not share that label. (This is implemented using integer
           pivoting)
        3. A label will now be duplicated in the other polytope, drop it in a
           similar way.
        4. Repeat steps 2 and 3 until have Nash Equilibrium.

        Parameters
        ----------

            initial_dropped_label: int

        Returns
        -------

            equilibria: A tuple.
        """
        return lemke_howson(
            *self.payoff_matrices, initial_dropped_label=initial_dropped_label
        )

    def fictitious_play(self, iterations, play_counts=None):
        """
        Return a given sequence of actions through fictitious play. The
        implementation corresponds to the description of chapter 2 of
        [Fudenberg1998]_.

        1. Players have a belief of the strategy of the other player: a vector
        representing the number of times the player has chosen a given strategy.
        2. Players choose a best response to the belief.
        3. Players update their belief based on the latest choice of the
        opponent.

        Parameters
        ----------

            iterations: int
            play_counts: iterator

        Returns
        -------

            plays: A generator
        """
        return fictitious_play(
            *self.payoff_matrices,
            iterations=iterations,
            play_counts=play_counts
        )

    def stochastic_fictitious_play(
        self, iterations, play_counts=None, etha=10 ** -1, epsilon_bar=10 ** -2
    ):
        """Return a given sequence of actions and mixed strategies through stochastic fictitious play. The
        implementation corresponds to the description given in [Hofbauer2002]_.

        Parameters
        ----------
            iterations: int
            play_counts: iterator
            etha: float
            epsilon_bar: float
        Returns
        -------
            plays: A generator
        """
        return stochastic_fictitious_play(
            *self.payoff_matrices,
            iterations=iterations,
            play_counts=play_counts,
            etha=etha,
            epsilon_bar=epsilon_bar
        )

    def replicator_dynamics(self, y0=None, timepoints=None):
        """
        Implement replicator dynamics
        Return an array showing probability of each strategy being played over
        time.
        The total population is constant. Strategies can either stay constant
        if equilibria is achieved, replicate or die.

        Parameters
        ----------
            A: nxm array, where n=m
            y0: array
            timepoints: array
        Returns
        -------
            xs: array
        """
        A, _ = self.payoff_matrices
        return replicator_dynamics(A=A, y0=y0, timepoints=timepoints)

    def asymmetric_replicator_dynamics(self, x0=None, y0=None, timepoints=None):
        """
        Returns two arrays, corresponding to the two players, showing the
        probability of each strategy being played over time using the asymmetric
        replicator dynamics algorithm.

        Parameters
        ----------
            x0 : array, optional
            y0 : array, optional
            timepoints : array, optional

        Returns
        -------
            xs1 : array
            xs2 : array
        """
        A, B = self.payoff_matrices
        return asymmetric_replicator_dynamics(
            A=A, B=B, x0=x0, y0=y0, timepoints=timepoints
        )



"""A class for the Lemke Howson algorithm with lexicographical ordering"""
from itertools import cycle

import numpy as np

from nashpy.integer_pivoting import make_tableau, pivot_tableau_lex

from .lemke_howson import shift_tableau, tableau_to_strategy


def lemke_howson_lex(A, B, initial_dropped_label=0):
    """
     Obtain the Nash equilibria using the Lemke Howson algorithm implemented
     using lexicographical integer pivoting. (Able to solve degenerate games)

     1. Start at the artificial equilibrium (which is fully labeled)
     2. Choose an initial label to drop and move in the polytope for which
        the vertex has that label to the edge that does not share that label.
    (This is implemented using integer pivoting and the choice of label
    to drop is implemented using lexicographical ordering)
     3. A label will now be duplicated in the other polytope, drop it in a
        similar way.
     4. Repeat steps 2 and 3 until have Nash Equilibrium.

     Parameters
     ----------

             initial_dropped_label: int

     Returns
     -------

             equilibria: A tuple.
    """

    if np.min(A) <= 0:
        A = A + abs(np.min(A)) + 1
    if np.min(B) <= 0:
        B = B + abs(np.min(B)) + 1

    # build tableaux
    col_tableau = make_tableau(A)
    col_tableau = shift_tableau(col_tableau, A.shape)
    row_tableau = make_tableau(B.transpose())
    full_labels = set(range(sum(A.shape)))

    # slack variables
    row_slack_variables = range(B.shape[0], sum(B.shape))
    col_slack_variables = range(A.shape[0])

    # non-basic variables
    row_non_basic_variables = full_labels - set(row_slack_variables)
    col_non_basic_variables = full_labels - set(col_slack_variables)

    # print(initial_dropped_label)
    if initial_dropped_label in row_non_basic_variables:
        tableaux = cycle(
            (
                (row_tableau, row_slack_variables, row_non_basic_variables),
                (col_tableau, col_slack_variables, col_non_basic_variables),
            )
        )
    else:
        tableaux = cycle(
            (
                (col_tableau, col_slack_variables, col_non_basic_variables),
                (row_tableau, row_slack_variables, row_non_basic_variables),
            )
        )

    # First pivot (to drop a label)
    next_tableau, next_slack_variables, next_non_basic_variables = next(
        tableaux
    )

    entering_label = pivot_tableau_lex(
        next_tableau,
        initial_dropped_label,
        next_slack_variables,
        next_non_basic_variables,
    )

    # keeps track of each tableau's non-basic variables
    next_non_basic_variables.add(entering_label)
    next_non_basic_variables.remove(initial_dropped_label)

    while col_non_basic_variables.union(row_non_basic_variables) != full_labels:
        next_tableau, next_slack_variables, next_non_basic_variables = next(
            tableaux
        )

        # the first label is 'entering' in the sense that it will enter the next
        # tableau's set of basic variables
        just_entered_label = entering_label
        entering_label = pivot_tableau_lex(
            next_tableau,
            entering_label,
            next_slack_variables,
            next_non_basic_variables,
        )

        next_non_basic_variables.add(entering_label)
        next_non_basic_variables.remove(just_entered_label)

    row_strategy = tableau_to_strategy(
        row_tableau,
        full_labels - row_non_basic_variables,
        range(A.shape[0]),
    )

    col_strategy = tableau_to_strategy(
        col_tableau,
        full_labels - col_non_basic_variables,
        range(A.shape[0], sum(A.shape)),
    )

    return row_strategy, col_strategy



"""A class for a normal form game"""
import numpy as np
from scipy.optimize import linprog
from scipy.spatial import HalfspaceIntersection


def build_halfspaces(M):
    """
    Build a matrix representation for a halfspace corresponding to:

        Mx <= 1 and x >= 0

    This is of the form:

       [M: -1]
       [-1: 0]

    As specified in
    https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.spatial.HalfspaceIntersection.html

    Parameters
    ----------

        M: a numpy array

    Returns:
    --------

        Numpy array
    """
    number_of_strategies, dimension = M.shape
    b = np.append(-np.ones(number_of_strategies), np.zeros(dimension))
    M = np.append(M, -np.eye(dimension), axis=0)
    halfspaces = np.column_stack((M, b.transpose()))
    return halfspaces


def find_feasible_point(halfspaces):
    """
    Use linear programming to find a point inside the halfspaces (needed to
    define it).

    Code taken from scipy documentation:
    https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.spatial.HalfspaceIntersection.html

    Parameters
    ----------

        halfspaces: a matrix representation of halfspaces

    Returns:
    --------

        numpy array
    """
    norm_vector = np.reshape(
        np.linalg.norm(halfspaces[:, :-1], axis=1), (halfspaces.shape[0], 1)
    )
    c = np.zeros((halfspaces.shape[1],))
    c[-1] = -1
    A = np.hstack((halfspaces[:, :-1], norm_vector))
    b = -halfspaces[:, -1:]
    res = linprog(c, A_ub=A, b_ub=b)
    return res.x[:-1]


def labels(vertex, halfspaces):
    """
    Return the labels of the facets on which lie a given vertex. This is
    calculated by carrying out the matrix multiplictation.

    Parameters
    ----------

        vertex: a numpy array
        halfspaces: a numpy array

    Returns
    -------

       set
    """
    b = halfspaces[:, -1]
    M = halfspaces[:, :-1]
    return set(np.where(np.isclose(np.dot(M, vertex), -b))[0])


def non_trivial_vertices(halfspaces):
    """
    Returns all vertex, label pairs (ignoring the origin).

    Parameters:

        halfspaces: a numpy array

    Returns:

        generator
    """
    feasible_point = find_feasible_point(halfspaces)
    hs = HalfspaceIntersection(halfspaces, feasible_point)
    hs.close()
    return (
        (v, labels(v, halfspaces))
        for v in hs.intersections
        if not np.all(np.isclose(v, 0)) and max(v) < np.inf
    )



from setuptools import find_packages, setup

# Read in the version number
exec(open("src/nashpy/version.py", "r").read())

with open("requirements.txt") as f:
    requirements = []
    for library in f.read().splitlines():
        requirements.append(library)


setup(
    name="nashpy",
    version=__version__,
    install_requires=requirements,
    author="Vince Knight, James Campbell",
    author_email=("knightva@cardiff.ac.uk"),
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="",
    license="The MIT License (MIT)",
    description="A library to compute equilibria of 2 player normal form games",
)



"""Code to carry out replicator dynamics"""
import numpy as np

# from scipy.integrate import solve_ivp (will change program to use solve_ivp later)
from scipy.integrate import odeint


def get_derivative_of_fitness(x, t, A):
    """
    Find the derivative of fitness function
    """
    f = np.dot(A, x)
    phi = np.dot(f, x)
    return x * (f - phi)


def replicator_dynamics(A, y0=None, timepoints=None):
    """
    Implement replicator dynamics
    """

    if timepoints is None:
        timepoints = np.linspace(0, 10, 1000)

    if y0 is None:
        number_of_strategies = len(A)
        y0 = np.ones(number_of_strategies) / number_of_strategies

    xs = odeint(func=get_derivative_of_fitness, y0=y0, t=timepoints, args=(A,))
    return xs


def get_derivative_of_asymmetric_fitness(x, t, A, B):
    """
    Find the derivative of fitness function for the asymmetric replicator
    dynamics scenario
    """
    number_of_rows = A.shape[0]
    row_vector = x[:number_of_rows]
    col_vector = x[number_of_rows:]

    f1 = A @ col_vector
    f2 = row_vector @ B

    phi1 = f1 @ row_vector
    phi2 = f2 @ col_vector

    row_derivative = row_vector * (f1 - phi1)
    col_derivative = col_vector * (f2 - phi2)

    return np.concatenate((row_derivative, col_derivative))


def asymmetric_replicator_dynamics(A, B, x0=None, y0=None, timepoints=None):
    """
    Implement asymmetric replicator dynamics
    """
    if timepoints is None:
        timepoints = np.linspace(0, 10, 1000)

    if x0 is None:
        number_of_strategies_A = A.shape[0]
        x0 = np.ones(number_of_strategies_A) / number_of_strategies_A

    if y0 is None:
        number_of_strategies_B = A.shape[1]
        y0 = np.ones(number_of_strategies_B) / number_of_strategies_B

    initial_values = np.concatenate((x0, y0))
    xs = odeint(
        func=get_derivative_of_asymmetric_fitness,
        y0=initial_values,
        t=timepoints,
        args=(
            A,
            B,
        ),
    )
    xs1 = xs[:, : A.shape[0]]
    xs2 = xs[:, A.shape[0] :]
    return xs1, xs2



"""Code to carry out stochastic fictitious learning"""
import numpy as np
from nashpy.learning.fictitious_play import update_play_count


def get_distribution_response_to_play_count(A, play_count, epsilon_bar, etha):
    """Obtain a mixed strategy as a probability distribution as a response to a given play count"""
    if np.sum(play_count) == 0:
        strategies = play_count + 1 / len(play_count)
    else:
        strategies = play_count / np.sum(play_count)
    utilities = A @ strategies
    noisy_utilities = utilities + np.random.random(A.shape[0]) * epsilon_bar
    logit_choice = np.exp(etha ** -1 * noisy_utilities) / np.sum(
        np.exp(etha ** -1 * noisy_utilities)
    )
    return logit_choice


def stochastic_fictitious_play(
    A, B, iterations, etha=10 ** -1, epsilon_bar=10 ** -2, play_counts=None
):
    """
    Implement stochastic_fictitious play
    """
    if play_counts is None:
        play_counts = [
            np.array([0 for _ in range(dimension)]) for dimension in A.shape
        ]

    distributions = None, None

    yield play_counts, distributions

    for repetition in range(iterations):

        distributions = [
            get_distribution_response_to_play_count(
                A=matrix,
                play_count=play_count,
                etha=etha,
                epsilon_bar=epsilon_bar,
            )
            for matrix, play_count in zip((A, B.transpose()), play_counts[::-1])
        ]

        plays = [
            np.random.choice(range(len(distribution)), p=distribution)
            for distribution in distributions
        ]

        play_counts = [
            update_play_count(play_count, play)
            for play_count, play in zip(play_counts, plays)
        ]
        yield play_counts, distributions
